from typing import List, Dict, Any
import numpy as np
import logging
import io
import json
from csle_common.dao.simulation_config.simulation_env_input_config import SimulationEnvInputConfig


class IntrusionRecoveryPomdpConfig(SimulationEnvInputConfig):
    """
    DTO containing the configuration of an intrusion recovery POMDP
    """

    def __init__(self, eta: float, p_a: float, p_c_1: float, p_c_2: float, p_u: float, BTR: int, negate_costs: bool,
                 seed: int, discount_factor: float, states: List[int], actions: List[int], observations: List[int],
                 cost_tensor: List[List[float]], observation_tensor: List[List[float]],
                 transition_tensor: List[List[List[float]]], b1: List[float], T: int, simulation_env_name: str,
                 gym_env_name: str, max_horizon: float = np.inf) -> None:
        """
        Initializes the DTO

        :param eta: the scaling factor for the cost function
        :param p_a: the intrusion probability
        :param p_c_1: the crash probability in the healthy state
        :param p_c_2: the crash probability in the compromised state
        :param p_u: the software upgrade probability
        :param BTR: the periodic recovery interval
        :param negate_costs: boolean flag indicating whether costs should be negated or not
        :param seed: the random seed
        :param discount_factor: the discount factor
        :param states: the list of states
        :param actions: the list of actions
        :param observations: the list of observations
        :param cost_tensor: the cost tensor
        :param observation_tensor: the observation tensor
        :param transition_tensor: the transition tensor
        :param b1: the initial belief
        :param T: the time horizon
        :param simulation_env_name: name of the simulation environment
        :param gym_env_name: name of the gym environment
        :param max_horizon: the maximum horizon to avoid infinie simulations
        """
        self.eta = eta
        self.p_a = p_a
        self.p_c_1 = p_c_1
        self.p_c_2 = p_c_2
        self.p_u = p_u
        self.BTR = BTR
        self.negate_costs = negate_costs
        self.seed = seed
        self.discount_factor = discount_factor
        self.states = states
        self.actions = actions
        self.observations = observations
        self.cost_tensor = cost_tensor
        self.observation_tensor = observation_tensor
        self.transition_tensor = transition_tensor
        self.b1 = b1
        self.T = T
        self.simulation_env_name = simulation_env_name
        self.gym_env_name = gym_env_name
        self.max_horizon = max_horizon

    def __str__(self) -> str:
        """
        :return: a string representation of the DTO
        """
        return (f"eta: {self.eta}, p_a: {self.p_a}, p_c_1: {self.p_c_1}, p_c_2: {self.p_c_2}, p_u: {self.p_u}, "
                f"BTR: {self.BTR}, negate_costs: {self.negate_costs}, seed: {self.seed}, "
                f"discount_factor: {self.discount_factor}, states: {self.states}, actions: {self.actions}, "
                f"observations: {self.observation_tensor}, cost_tensor: {self.cost_tensor}, "
                f"observation_tensor: {self.observation_tensor}, transition_tensor: {self.transition_tensor}, "
                f"b1:{self.b1}, T: {self.T}, simulation_env_name: {self.simulation_env_name}, "
                f"gym_env_name: {self.gym_env_name}, max_horizon: {self.max_horizon}")

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "IntrusionRecoveryPomdpConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        dto = IntrusionRecoveryPomdpConfig(
            eta=d["eta"], p_a=d["p_a"], p_c_1=d["p_c_1"], p_c_2=d["p_c_2"], p_u=d["p_u"], BTR=d["BTR"],
            negate_costs=d["negate_costs"], seed=d["seed"], discount_factor=d["discount_factor"], states=d["states"],
            actions=d["actions"], observations=d["observations"], cost_tensor=d["cost_tensor"],
            observation_tensor=d["observation_tensor"], transition_tensor=d["transition_tensor"], b1=d["b1"],
            T=d["T"], simulation_env_name=d["simulation_env_name"], gym_env_name=d["gym_env_name"])
        return dto

    def to_dict(self) -> Dict[str, Any]:
        """
        Gets a dict representation of the object

        :return: A dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["eta"] = self.eta
        d["p_a"] = self.p_a
        d["p_c_1"] = self.p_c_1
        d["p_c_2"] = self.p_c_2
        d["p_u"] = self.p_u
        d["BTR"] = self.BTR
        d["negate_costs"] = self.negate_costs
        d["seed"] = self.seed
        d["discount_factor"] = self.discount_factor
        d["states"] = self.states
        d["actions"] = self.actions
        d["observations"] = self.observations
        d["cost_tensor"] = self.cost_tensor
        d["observation_tensor"] = self.observation_tensor
        d["transition_tensor"] = self.transition_tensor
        d["b1"] = self.b1
        d["T"] = self.T
        d["simulation_env_name"] = self.simulation_env_name
        d["gym_env_name"] = self.gym_env_name
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "IntrusionRecoveryPomdpConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        logging.info(msg="msg")
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
            print(json_str)
        return IntrusionRecoveryPomdpConfig.from_dict(json.loads(json_str))



from typing import List
import math
import numpy as np


class GeneralUtil:
    """
    Class with general utility functions related to csle-tolerance
    """

    @staticmethod
    def threshold_probability(b1: float, threshold: float, k=-20) -> float:
        """
        Returns the probability of taking an action given a belief and a threshold

        :param b1: the belief
        :param threshold: the threshold
        :return: the stopping probability
        """
        if (1 - round(b1, 2)) == 0:
            return 1.0
        if round(b1, 2) == 0:
            return 0.0
        if (threshold * (1 - b1)) > 0 and (b1 * (1 - threshold)) / (
            threshold * (1 - b1)
        ) > 0:
            try:
                return math.pow(
                    1 + math.pow(((b1 * (1 - threshold)) / (threshold * (1 - b1))), k),
                    -1,
                )
            except Exception:
                return 0.0
        else:
            return 0.0

    @staticmethod
    def sigmoid(x) -> float:
        """
        The sigmoid function

        :param x: the input
        :return: sigmoid(x)
        """
        return 1 / (1 + math.exp(-x))

    @staticmethod
    def inverse_sigmoid(y) -> float:
        """
        The inverse sigmoid function

        :param y: sigmoid(x)
        :return: sigmoid(x)^(-1)
        """
        return math.log(y / (1 - y), math.e)

    @staticmethod
    def sample_next_state(
        transition_tensor: List[List[List[float]]], s: int, a: int, states: List[int]
    ) -> int:
        """
        Samples the next state of a MDP or POMDP

        :param transition_tensor: the transition tensor
        :param s: the current state
        :param a: the current action
        :param states: the list of states
        :return: the next state
        """
        state_probs = []
        for s_prime in states:
            state_probs.append(transition_tensor[a][s][s_prime])
        return int(np.random.choice(np.arange(0, len(states)), p=state_probs))

    @staticmethod
    def register_envs() -> None:
        """
        Utility method for registering Gymnasium environments

        :return: None
        """
        from gymnasium.envs.registration import register

        register(
            id="csle-tolerance-intrusion-recovery-pomdp-v1",
            entry_point="csle_tolerance.envs.intrusion_recovery_pomdp_env:IntrusionRecoveryPomdpEnv",
            kwargs={"config": None},
        )
        register(
            id="csle-tolerance-intrusion-response-cmdp-v1",
            entry_point="csle_tolerance.envs.intrusion_response_cmdp_env:IntrusionResponseCmdpEnv",
            kwargs={"config": None},
        )



from csle_tolerance.util.intrusion_response_cmdp_util import IntrusionResponseCmdpUtil


class TestInstrusionToleranceCmdpSuite:
    """
    Test suite for intrusion_response_cmdp_util.py
    """

    def test_state_space(self) -> None:
        """
        Tests the state space of the CMDP

        :return: None
        """
        assert len(IntrusionResponseCmdpUtil.state_space(2)) != 0
        assert IntrusionResponseCmdpUtil.state_space(1) is not None
        assert IntrusionResponseCmdpUtil.state_space(0) == [0]
        assert all(isinstance(n, int) for n in IntrusionResponseCmdpUtil.state_space(3))

    def test_action_space(self) -> None:
        """
        Tests the action space of the CMDP

        :return: None
        """
        assert IntrusionResponseCmdpUtil.action_space() == [0, 1]

    def test_cost_function(self) -> None:
        """
        Tests the cost function of the CMDP

        :return: None
        """
        s = 1
        assert IntrusionResponseCmdpUtil.cost_function(s, True) == -1.0
        assert IntrusionResponseCmdpUtil.cost_function(s, False) == 1.0
        assert isinstance(IntrusionResponseCmdpUtil.cost_function(s, False), float)
        assert isinstance(IntrusionResponseCmdpUtil.cost_function(s, True), float)

    def test_cost_tensor(self) -> None:
        """
        Tests the function of creating a tensor with the costs of the CMDP

        :return: None
        """
        states = [0, 1, 2, 3]
        expected = [-0.0, -1.0, -2.0, -3.0]
        assert IntrusionResponseCmdpUtil.cost_tensor(states, True) == expected
        assert IntrusionResponseCmdpUtil.cost_tensor(states, False) == [
            -item for item in expected
        ]

    def test_constraint_cost_function(self) -> None:
        """
        Tests the constraint cost function of the CMDP

        :return: None
        """
        states = [1, 2]
        f = 5
        expected = [0.0, 0.0]
        assert (
            IntrusionResponseCmdpUtil.constraint_cost_function(states[0], f)
            == expected[0]
        )
        assert (
            IntrusionResponseCmdpUtil.constraint_cost_function(states[1], f)
            == expected[1]
        )

    def test_constraint_cost_tensor(self) -> None:
        """
        Tests the function of creating a tensor with the constrained costs of the CMDP

        :return: None
        """
        states = [1, 2]
        f = 5
        expected = [0.0, 0.0]
        assert IntrusionResponseCmdpUtil.constraint_cost_tensor(states, f) == expected

    def test_delta_function(self) -> None:
        """
        Tests the delta function that gives the probability of the change in the number of the healthy nodes

        :return: None
        """
        s = 1
        s_max = 2
        p_a = 0.2
        p_c = 0.5
        p_u = 0.7
        delta = 1
        expected = 1.35
        assert (
            round(
                IntrusionResponseCmdpUtil.delta_function(
                    s, p_a, p_c, p_u, delta, s_max
                ),
                2,
            )
            == expected
        )

    def test_transition_function(self) -> None:
        """
        Tests the transition function of the CMDP

        :return: None
        """
        s = 0
        s_prime = 1
        a = 1
        s_max = 2
        p_a = 0.2
        p_c = 0.3
        p_u = 0.5
        delta = s_prime - a - s
        assert (
            IntrusionResponseCmdpUtil.delta_function(s, p_a, p_c, p_u, delta, s_max)
            == 3
        )

    def test_transition_tensor(self) -> None:
        """
        Tests the transition tensor function of the CMDP

        :return: None
        """
        states = [0, 1]
        actions = [0]
        s_max = 2
        p_a = 0.2
        p_c = 0.3
        p_u = 0.5
        expected = [[[3 / 5, 2 / 5], [2 / 5, 3 / 5]]]
        transition_tensor = IntrusionResponseCmdpUtil.transition_tensor(
            states, actions, p_a, p_c, p_u, s_max
        )
        for i in range(len(transition_tensor)):
            for j in range(len(transition_tensor[i])):
                for k in range(len(transition_tensor[i][j])):
                    transition_tensor[i][j][k] = round(transition_tensor[i][j][k], 1)
        assert transition_tensor == expected



from csle_tolerance.util.pomdp_solve_parser import PomdpSolveParser
import pytest
import tempfile
import numpy as np


class TestInstrusionTolerancePomdpSolveParserSuite:
    """
    Test suite for pomdp_solve_parser.py
    """

    def test_parse_alpha_vectors(self) -> None:
        """
        Tests the function of parsing alpha vectors from a given file location

        :return: None
        """
        # test for a scenario where the file does not exist
        with pytest.raises(FileNotFoundError):
            file_path = "file.txt"
            PomdpSolveParser.parse_alpha_vectors(file_path)
        # test case for correct parsing
        file_content = "0\n0.1 0.2 0.3\n1\n0.4 0.5 0.6"
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(file_content)
            f.flush()
            file_path = f.name
            result = PomdpSolveParser.parse_alpha_vectors(file_path)
            expected = [(0, [0.1, 0.2, 0.3]), (1, [0.4, 0.5, 0.6])]
            assert result == expected
        # test case for incorrect file format
        file_content = "Invalid content"
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(file_content)
            f.flush()
            file_path = f.name
            result = PomdpSolveParser.parse_alpha_vectors(file_path)
            assert result == []

    def test_optimal_avg_value(self) -> None:
        """
        Tests the function of computing the optimal average value given a set of alpha vectors

        :return: None
        """
        alpha_vectors_content = "0\n0.1 0.2 0.3\n1\n0.4 0.5 0.6"
        initial_belief = 0.5
        btr = 10
        with tempfile.NamedTemporaryFile(mode="w") as f:
            f.write(alpha_vectors_content)
            f.flush()
            result = PomdpSolveParser.optimal_avg_value(f.name, initial_belief, btr)
        expected = -0.45 / btr
        assert np.isclose(result, expected)
